import csv
import json
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar, Union

if TYPE_CHECKING:
    from policy_inspector.models import MainModel

logger = logging.getLogger(__name__)

ModelClass = TypeVar("ModelClass", bound="MainModel")
"""Type variable for model classes derived from MainModel."""


def load_json(
    file_path: Path,
    encoding: str = "utf-8",
) -> Union[list[dict], Any]:
    """Loads JSON file from given file_path and return it's content."""
    return json.loads(file_path.read_text(encoding=encoding))


def load_csv(
    file_path: Path,
    encoding: str = "utf-8",
) -> Union[list[dict], Any]:
    """Loads CSV file from given file_path and return it's content."""
    return list(csv.DictReader(file_path.open(encoding=encoding)))


class Loader:
    _loaders = {"json": load_json, "csv": load_csv}
    """Mapping of file extensions to example loading functions."""
    parser_suffix: str = "parse_"

    @classmethod
    def load_model(
        cls, model_cls: type[ModelClass], file_path: Path
    ) -> list[ModelClass]:
        """Load given file and create instances of the specified model class.

        Args:
            model_cls: The model class to instantiate for each example entry.
            file_path: The path to the JSON or CSV file containing the example.

        Returns:
            A list of instances of the specified model class.
        """
        ext = file_path.suffix.lower().lstrip(".")

        if ext not in cls._loaders:
            raise ValueError(f"Unsupported file type: {ext}")

        loader = cls._loaders[ext]
        raw_items = loader(file_path)

        parser_name = f"{cls.parser_suffix}{ext}"
        parser_method = getattr(model_cls, parser_name, None)
        if parser_method is None:
            raise ValueError(f"{model_cls.__name__} lacks {parser_name} method")

        return [parser_method(item) for item in raw_items]



import logging
from ipaddress import IPv4Network
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from policy_inspector.models import AddressGroup, AddressObject

logger = logging.getLogger(__name__)


class AddressResolver:
    """Process type of class to resolve Address Groups and Address Objects into their actual IP addresses.

    It expands Address Groups (AG) recursively and converts Address Objects (AO) into IP networks.

    Args:
        address_objects: A list of ``AddressObject``.
        address_groups: A list of ``AddressGroup``.
    """

    def __init__(
        self,
        address_objects: list["AddressObject"],
        address_groups: list["AddressGroup"],
    ):
        self.address_objects: dict[str, str] = {
            ao.name: ao.ip_netmask for ao in address_objects
        }
        self.address_groups: dict[str, set[str]] = {
            ag.name: ag.static for ag in address_groups
        }
        self.cache: dict[str, set[IPv4Network]] = {}

    def resolve(self, names: set[str]) -> set[IPv4Network]:
        """Resolve given names of ``Address Groups`` or ``Address Objects`` to actual IP address.

        Args:
            names: Names of ``Address Groups`` or ``Address Objects``

        Returns:
            Set of ``IPv4Network`` of all ``names``.

        """
        result = set()
        for name in names:
            result.update(self._resolve_name(name))
        return result

    def _resolve_name(self, name: str) -> set[IPv4Network]:
        """Resolve single ``name``"""
        if name in self.cache:
            return self.cache[name]

        if name in self.address_groups:
            logger.debug(f"Resolving group: {name}")
            resolved = set()
            for member in self.address_groups[name]:
                resolved.update(self._resolve_name(member))
            self.cache[name] = resolved
            return resolved

        if name in self.address_objects:
            logger.debug(f"Resolving object: {name}")
            try:
                network = IPv4Network(self.address_objects[name], strict=False)
                result = {network}
                self.cache[name] = result
                return result
            except ValueError as e:
                raise ValueError(f"Invalid IP network for {name}: {e}") from e

        raise ValueError(f"Unknown address object/group: {name}")



from typing import TYPE_CHECKING

from rich.table import Table

if TYPE_CHECKING:
    from policy_inspector.models import SecurityRule


def display_rules(*rules: "SecurityRule", title: str = "Rules"):
    table = Table(title=title)

    main_headers = ("Attribute", "Security Rule")
    next_headers = [f"Preceding Rule {i}" for i in range(len(rules) - 1)]
    map(table.add_column, main_headers, next_headers)

    for attribute in SecurityRule.model_fields.keys():
        rules_attribute = [getattr(rule, attribute) for rule in rules]
        table.add_row(attribute, *rules_attribute)

    print(table)



# ruff: noqa: RET503
import logging
from pathlib import Path
from typing import Any, Callable, Optional

from click import Context, Parameter, UsageError, option
from click.types import Choice as clickChoice
from pydantic import BaseModel
from rich.logging import RichHandler

EXAMPLES_DIR = Path(__file__).parent / "example"


def get_example_file_path(file_path: Path) -> Path:
    return EXAMPLES_DIR / file_path


# It's just make use of pydantic because it's available ;)
class Example(BaseModel):
    name: str
    args: list
    cmd: Callable

    def model_post_init(self, data):
        self.args = [get_example_file_path(arg) for arg in self.args]


def verbose_option(logger) -> Callable:
    """Wrapper around Click ``option``. Sets logger and its handlers to the ``DEBUG`` level."""

    def callback(ctx, param, value) -> None:
        if not value:
            return
        count = len(value)
        if count > 0:
            logger.setLevel(logging.DEBUG)
            for handler in logger.handlers:
                handler.setLevel(logging.DEBUG)
        if count > 1:
            handler: RichHandler = logger.handlers[0]
            handler.enable_link_path = True
            handler._log_render.show_path = True
            handler._log_render.show_time = True
            handler._log_render.show_level = True

    kwargs = {
        "is_flag": True,
        "multiple": True,
        "callback": callback,
        "expose_value": False,
        "is_eager": True,
        "help": "Set log messages to DEBUG",
    }
    return option("-v", "--verbose", **kwargs)


def config_logger(
    logger: logging.Logger,
    level: str = "INFO",
    log_format: str = "%(message)s",
    date_format: str = "[%X]",
) -> None:
    """
    Configure ``logger`` with ``RichHandler``

    Args:
        logger: Instance of a ``logging.Logger``
    """
    logger.setLevel(level)
    rich_handler = RichHandler(
        level=level,
        rich_tracebacks=True,
        # tracebacks_suppress=[click],
        show_path=False,
        show_time=False,
        show_level=False,
        omit_repeated_times=False,
    )
    formatter = logging.Formatter(log_format, date_format, "%")
    rich_handler.setFormatter(formatter)
    logger.handlers = [rich_handler]


class ExampleChoice(clickChoice):
    def __init__(self, examples: list[Example]) -> None:
        self.examples = {example.name: example for example in examples}
        super().__init__(list(self.examples.keys()), False)  # noqa: FBT003

    def convert(
        self, value: Any, param: Optional["Parameter"], ctx: Optional["Context"]
    ) -> Any:
        normed_value = value
        normed_choices = self.examples

        if ctx is not None and ctx.token_normalize_func is not None:
            normed_value = ctx.token_normalize_func(value)
            normed_choices = {
                ctx.token_normalize_func(normed_choice): original
                for normed_choice, original in normed_choices.items()
            }

        normed_value = normed_value.casefold()
        normed_choices = {
            normed_choice.casefold(): original
            for normed_choice, original in normed_choices.items()
        }

        try:
            return normed_choices[normed_value]
        except KeyError:
            matching_choices = list(
                filter(lambda c: c.startswith(normed_value), normed_choices)
            )

        if len(matching_choices) == 1:
            return matching_choices[0]

        if not matching_choices:
            choices_str = ", ".join(map(repr, self.choices))
            message = f"{value!r} is not one of {choices_str}."
        else:
            choices_str = ", ".join(map(repr, matching_choices))
            message = f"{value!r} too many matches: {choices_str}."
        raise UsageError(message=message, ctx=ctx)



from collections.abc import Iterable, Iterator
from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from policy_inspector.models import SecurityRule

FilterFunction = Callable[["SecurityRule"], bool]


def apply_filters(
    filters: Iterable[FilterFunction],
    policies: Iterable["SecurityRule"],
) -> Iterator["SecurityRule"]:
    """Apply a set of filter functions to security policies.

    Args:
        filters: An iterable of filter functions that take a `SecurityRule` object
            and return a boolean indicating whether the rule should be included.
        policies: An iterable of `SecurityRule` objects to be filtered.

    Returns:
        An iterator over `SecurityRule` objects that pass all filter conditions.

    Example:
        filters = [exclude_disabled, exclude_deny]
        filtered_policies = apply_filters(filters, policies)
        for policy in filtered_policies:
            print(policy)

    """
    return filter(
        lambda p: all(filter_func(p) for filter_func in filters),
        policies,
    )


def exclude_disabled(policy: "SecurityRule") -> bool:
    """Exclude security rules that are disabled."""
    return True if policy.enabled else False


def exclude_deny(policy: "SecurityRule") -> bool:
    """Exclude security rules with a 'deny' action."""
    return policy.action != "deny"



import logging
from pathlib import Path

import rich_click as click
from click import Path as ClickPath
from click import UsageError

from policy_inspector.loader import Loader, ModelClass
from policy_inspector.models import (
    AddressGroup,
    AddressObject,
    SecurityRule,
)
from policy_inspector.scenario import Scenario
from policy_inspector.scenario.shadowing import Shadowing
from policy_inspector.scenario.shadowing_by_value import ShadowingByValue
from policy_inspector.utils import (
    Example,
    ExampleChoice,
    config_logger,
    verbose_option,
)

logger = logging.getLogger()
config_logger(logger)
click.rich_click.SHOW_ARGUMENTS = True
click.rich_click.TEXT_MARKUP = "markdown"
# click.rich_click.APPEND_METAVARS_HELP = True
click.rich_click.SHOW_METAVARS_COLUMN = True


@click.group(no_args_is_help=True, add_help_option=True)
@verbose_option(logger)
def main():
    """*PINS*
    as Policy Inspector
    """


@main.command("list")
@verbose_option(logger)
def main_list() -> None:
    """List available Scenarios."""
    logger.info("Available Scenarios:")
    scenarios = Scenario.get_available()
    for name, scenario in scenarios.items():
        logger.info(f"- {name}")
        logger.debug(f"  {scenario.__doc__}")
        for check in scenario.checks:
            logger.debug(f"  - {check.__name__}")


@main.group("run", no_args_is_help=True)
@verbose_option(logger)
def main_run():
    """Execute Scenario."""


@main_run.command("shadowing", no_args_is_help=True)
@verbose_option(logger)
@click.argument(
    "security_rules_path",
    required=True,
    type=ClickPath(dir_okay=False, path_type=Path),
)
def run_shadowing(security_rules_path: Path) -> None:
    security_rules = load_model(SecurityRule, security_rules_path)
    scenario = Shadowing(security_rules)
    process(scenario)


@main_run.command("complex_shadowing", no_args_is_help=True)
@verbose_option(logger)
@click.argument(
    "security_rules_path",
    required=True,
    type=ClickPath(dir_okay=False, path_type=Path),
)
@click.argument(
    "address_groups_path",
    required=True,
    type=ClickPath(dir_okay=False, path_type=Path),
)
@click.argument(
    "address_objects_path",
    required=True,
    type=ClickPath(dir_okay=False, path_type=Path),
)
def run_complex_shadowing(
    security_rules_path: Path,
    address_groups_path: Path,
    address_objects_path: Path,
) -> None:
    security_rules = load_model(SecurityRule, security_rules_path)
    address_groups = load_model(AddressGroup, address_groups_path)
    address_objects = load_model(AddressObject, address_objects_path)
    scenario = ShadowingByValue(security_rules, address_objects, address_groups)
    process(scenario)


def process(scenario: Scenario):
    """Helper function for executing and analyzing a Scenario."""
    logger.info(f"→ Executing '{scenario.name}' scenario")
    try:
        output = scenario.execute()
        logger.info("")
        logger.info("▶ Results")
        logger.info("―――――――――")
        scenario.analyze(output)
    except Exception as ex:  # noqa: BLE001
        raise UsageError(str(ex))  # noqa: B904


def load_model(
    model_cls: type[ModelClass], file_path: Path
) -> list[ModelClass]:
    """Helper function for loading models from file."""
    logger.info(f"↺ Loading {model_cls.plural}")
    instances = Loader.load_model(model_cls, file_path)
    logger.debug(f"✓ Loaded {len(instances)} {model_cls.plural} successfully")
    return instances


examples = [
    Example(
        name="shadowing",
        args=[Path("1/policies.json")],
        cmd=run_shadowing,
    ),
    Example(
        name="shadowing_long",
        args=[Path("2/policies.json")],
        cmd=run_shadowing,
    ),
    Example(
        name="shadowing_with_addresses",
        args=[
            Path("1/policies.json"),
            Path("1/address_groups.json"),
            Path("1/address_objects.json"),
        ],
        cmd=run_complex_shadowing,
    ),
    Example(
        name="shadowing_with_addresses2",
        args=[
            Path("2/policies.json"),
            Path("2/address_groups.json"),
            Path("2/address_objects.json"),
        ],
        cmd=run_complex_shadowing,
    ),
]


@main_run.command("example", no_args_is_help=True)
@verbose_option(logger)
@click.argument(
    "example",
    type=ExampleChoice(examples),
)
@click.pass_context
def run_example(ctx, example: Example) -> None:
    """Run one of the examples."""
    logger.info(f"▶ Selected '{example.name}' example")
    ctx.invoke(example.cmd.callback, *example.args)


if __name__ == "__main__":
    main()



import logging
from typing import ClassVar, Literal, Optional, Union

from pydantic import BaseModel, Field
from pydantic.networks import IPv4Network

AnyObj = "any"
AnyObjType = set[Literal["any"]]

AppDefault = "application-default"
AppDefaultType = set[Literal["application-default"]]

SetStr = set[str]
Action = Literal["allow", "deny", "monitor"]

logger = logging.getLogger(__name__)


class MainModel(BaseModel):
    """Base class for all models."""

    singular: ClassVar[Optional[str]] = None
    """Display name of a single model."""
    plural: ClassVar[Optional[str]] = None
    """Display name of a many models."""


class SecurityRule(MainModel):
    singular: ClassVar[str] = "Security Rule"
    plural: ClassVar[str] = "Security Rules"

    name: str = Field(
        ...,
        description="Name of a rule.",
    )
    enabled: bool = Field(
        default=True,
        description="Whether the policy is enabled or disabled",
    )
    action: Action = Field(
        default="allow",
        description="Whether the traffic should be allowed or denied.",
    )
    source_zones: Union[SetStr, AnyObjType] = Field(
        default={"any"},
        description="Set of source zones or 'any'",
    )
    destination_zones: Union[SetStr, AnyObjType] = Field(
        default={"any"},
        description="Set of destination zones or 'any'",
    )

    source_addresses: Union[SetStr, AnyObjType] = Field(
        default={"any"},
        description="Source address objects/groups or 'any'",
    )

    destination_addresses: Union[SetStr, AnyObjType] = Field(
        default={"any"},
        description="Destination address objects/groups or 'any'",
    )

    applications: Union[SetStr, AnyObjType] = Field(
        default={"any"},
        description="Set of applications or 'any' that the rule applies to.",
    )

    services: Union[SetStr, AnyObjType, AppDefaultType] = Field(
        default_factory=set,
        description="Services (e.g., TCP/UDP ports) or 'any'/'application-default'",
    )

    category: Union[SetStr, AnyObjType] = Field(
        default={"any"},
        description="URL categories or 'any'",
    )

    source_addresses_ip: Optional[set[IPv4Network]] = Field(
        default=None,
        description="Resolved source addresses to a set of IP networks. ",
    )

    destination_addresses_ip: Optional[set[IPv4Network]] = Field(
        default=None,
        description="Resolved destination addresses to a set of IP networks.",
    )

    @classmethod
    def parse_json(cls, data: dict) -> "SecurityRule":
        """Map a JSON object to a SecurityRule."""
        mapping = {
            "@name": "name",
            "source": "source_addresses",
            "destination": "destination_addresses",
            "from": "source_zones",
            "to": "destination_zones",
            "application": "applications",
            "service": "services",
            "category": "category",
        }

        def extract_value(value):
            if isinstance(value, dict) and "member" in value:
                return set(value["member"])
            return value

        parsed = {mapping.get(k, k): extract_value(v) for k, v in data.items()}
        return cls(**parsed)

    @classmethod
    def parse_csv(cls, data: dict) -> "SecurityRule":
        """Map a CSV row to a SecurityRule."""
        mapping = {
            "Name": "name",
            "Source Address": "source_addresses",
            "Destination Address": "destination_addresses",
            "Source Zone": "source_zones",
            "Destination Zone": "destination_zones",
            "Application": "applications",
            "Service": "services",
            "Category": "category",
        }
        list_fields = {
            "source_addresses",
            "destination_addresses",
            "source_zones",
            "destination_zones",
            "applications",
            "services",
            "category",
        }
        parsed_data = {}

        for key, value in data.items():
            mapped_key = mapping.get(key, key)
            key_value = value
            if mapped_key in list_fields:
                key_value = set(value.split(";")) if value else set()
            parsed_data[mapped_key] = key_value

        return cls(**parsed_data)


class AddressGroup(MainModel):
    singular: ClassVar[str] = "Address Group"
    plural: ClassVar[str] = "Address Groups"

    name: str = Field(..., description="Name of the address group.")
    description: str = Field(default="")
    tag: set[str] = Field(default_factory=set)
    static: set[str] = Field(default_factory=set)

    @classmethod
    def parse_json(cls, data: dict) -> "AddressGroup":
        """Map a JSON object to an AddressGroup."""
        mapping = {"@name": "name", "ip-address": "ip_address"}
        list_fields = ("tag", "static")

        parsed = {}
        for key, value in data.items():
            mapped_key = mapping.get(key, key)
            key_value = value
            if value and mapped_key in list_fields:
                key_value = set(value) if value else set()
            parsed[mapped_key] = key_value
        return cls(**parsed)


class AddressObject(MainModel):
    singular: ClassVar[str] = "Address Object"
    plural: ClassVar[str] = "Address Objects"

    name: str = Field(..., description="Name of the address object.")
    ip_netmask: str = Field(
        ...,
        description="IP address and netmask of the address object.",
    )

    @classmethod
    def parse_json(cls, data: dict) -> "AddressObject":
        """Map a JSON object to an AddressObject."""
        mapping = {"@name": "name", "ip-netmask": "ip_netmask"}
        parsed = {mapping.get(k, k): v for k, v in data.items()}
        return cls(**parsed)



import logging
from typing import TYPE_CHECKING, Callable

from policy_inspector.models import AnyObj
from policy_inspector.scenario.base import CheckResult, Scenario

if TYPE_CHECKING:
    from policy_inspector.models import SecurityRule

logger = logging.getLogger(__name__)

ShadowingCheckFunction = Callable[["SecurityRule", "SecurityRule"], CheckResult]

ChecksOutputs = dict[str, CheckResult]
"""Dict with check's name as keys and its output as value."""

PrecedingRulesOutputs = dict[str, ChecksOutputs]
"""Dict with Preceding Rule's name as keys and ChecksOutputs as its value."""


def check_action(
    rule: "SecurityRule",
    preceding_rule: "SecurityRule",
) -> CheckResult:
    """Check if the action is the same in both rules."""
    result = rule.action == preceding_rule.action
    message = "Actions match" if result else "Actions differ"
    return result, message


def check_source_zone(
    rule: "SecurityRule",
    preceding_rule: "SecurityRule",
) -> CheckResult:
    """Checks the source zones of the preceding rule."""
    if rule.source_zones == preceding_rule.source_zones:
        return True, "Source zones are the same"

    if preceding_rule.source_zones.issubset(rule.source_zones):
        return True, "Preceding rule source zones cover rule's source zones"

    if AnyObj in preceding_rule.source_zones:
        return True, "Preceding rule source zones is 'any'"

    return False, "Source zones differ"


def check_destination_zone(
    rule: "SecurityRule",
    preceding_rule: "SecurityRule",
) -> CheckResult:
    """Checks the destination zones of the preceding rule."""
    if rule.destination_zones == preceding_rule.destination_zones:
        return True, "Destination zones are the same"

    if rule.destination_zones.issubset(preceding_rule.destination_zones):
        return (
            True,
            "Preceding rule destination zones cover rule's destination zones",
        )

    if AnyObj in preceding_rule.destination_zones:
        return True, "Preceding rule destination zones is 'any'"

    return False, "Destination zones differ"


def check_source_address(
    rule: "SecurityRule",
    preceding_rule: "SecurityRule",
) -> CheckResult:
    """Checks the source addresses of the preceding rule's addresses."""
    if rule.source_addresses == preceding_rule.source_addresses:
        return True, "Source addresses are the same"

    if AnyObj in preceding_rule.source_addresses:
        return True, "Preceding rule allows any source address"

    if AnyObj in rule.source_addresses:
        return False, "Rule not covered due to 'any' source"

    if rule.source_addresses.issubset(preceding_rule.source_addresses):
        return (
            True,
            "Preceding rule source addresses cover rule's source addresses",
        )

    return False, "Source addresses not covered at all"


def check_destination_address(
    rule: "SecurityRule",
    preceding_rule: "SecurityRule",
) -> CheckResult:
    """Checks if the destination addresses are
    identical, allow any, or are subsets of the preceding rule's addresses.
    """
    if AnyObj in preceding_rule.destination_addresses:
        return True, "Preceding rule allows any destination address"

    if rule.destination_addresses == preceding_rule.destination_addresses:
        return True, "Destination addresses are the same"

    if rule.destination_addresses.issubset(
        preceding_rule.destination_addresses,
    ):
        return (
            True,
            "Preceding rule destination addresses cover rule's destination addresses",
        )

    return False, "Destination addresses not covered at all"


def check_application(
    rule: "SecurityRule",
    preceding_rule: "SecurityRule",
) -> CheckResult:
    """Checks the applications of the preceding rule."""
    rule_apps = rule.applications
    preceding_apps = preceding_rule.applications

    if rule_apps == preceding_apps:
        return True, "The same applications"

    if AnyObj in preceding_apps:
        return True, "Preceding rule allows any application"

    if rule_apps.issubset(preceding_apps):
        return True, "Preceding rule contains rule's applications"

    return False, "Rule doesn't cover"


def check_services(
    rule: "SecurityRule",
    preceding_rule: "SecurityRule",
) -> CheckResult:
    """Checks if the rule's ports are the same
    or a subset of the preceding rule's ports.
    """
    if rule.services == preceding_rule.services:
        return True, "Preceding rule and rule's services are the same"

    if all(service in preceding_rule.services for service in rule.services):
        return True, "Preceding rule contains rule's applications"

    return False, "Preceding rule does not contain all rule's applications"


class Shadowing(Scenario):
    """
    This scenario identifies when a rule is completely shadowed by a
    preceding rule. Shadowing occurs when a rule will never be matched
    because a rule earlier in the processing order would always match
    first.
    """

    name: str = "Shadowing"
    checks: list[ShadowingCheckFunction] = [
        check_action,
        check_application,
        check_services,
        check_source_zone,
        check_destination_zone,
        check_source_address,
        check_destination_address,
    ]

    def __init__(self, security_rules: list["SecurityRule"]):
        self.security_rules = security_rules

    def execute(self) -> dict[str, PrecedingRulesOutputs]:
        rules = self.security_rules
        results = {}
        for i, rule in enumerate(rules):
            output = {}
            for j in range(i):
                preceding_rule = rules[j]
                output[preceding_rule.name] = self.run_checks(
                    rule,
                    preceding_rule,
                )
            results[rule.name] = output
        return results

    def analyze(
        self,
        results: dict[str, PrecedingRulesOutputs],
    ):
        for rule_name, rule_results in results.items():
            shadowing_rules = []
            for preceding_rule_name, checks_results in rule_results.items():
                if all(
                    check_result[0] for check_result in checks_results.values()
                ):
                    shadowing_rules.append(preceding_rule_name)
            if shadowing_rules:
                logger.info(f"✖ '{rule_name}' shadowed by:")
                for rule in shadowing_rules:
                    logger.info(f"   • {rule}")
            else:
                logger.debug(f"✔ '{rule_name}' not shadowed")



import logging
from typing import TYPE_CHECKING

from policy_inspector.models import AnyObj
from policy_inspector.resolver import AddressResolver
from policy_inspector.scenario.shadowing import (
    CheckResult,
    Shadowing,
    ShadowingCheckFunction,
    check_action,
    check_application,
    check_destination_zone,
    check_services,
    check_source_zone,
)

from .shadowing import PrecedingRulesOutputs

if TYPE_CHECKING:
    from policy_inspector.models import (
        AddressGroup,
        AddressObject,
        SecurityRule,
    )

logger = logging.getLogger(__name__)


def check_services_and_application(
    rule: "SecurityRule",
    preceding_rule: "SecurityRule",
) -> CheckResult:
    pass


def check_source_addresses_by_ip(
    rule: "SecurityRule",
    preceding_rule: "SecurityRule",
) -> CheckResult:
    if rule.source_addresses == preceding_rule.source_addresses:
        return True, "Source addresses are the same"

    if AnyObj in preceding_rule.source_addresses:
        return True, "Preceding rule allows any source address"

    if AnyObj in rule.source_addresses:
        return False, "Rule not covered due to 'any' source"

    for addr in rule.source_addresses_ip:
        if not any(
            addr.subnet_of(net) for net in preceding_rule.source_addresses_ip
        ):
            return (
                False,
                f"Source ip address {addr} is not covered by preceding rule",
            )

    return True, "Preceding rule covers all source ip addresses"


def check_destination_addresses_by_ip(
    rule: "SecurityRule",
    preceding_rule: "SecurityRule",
) -> CheckResult:
    if rule.destination_addresses == preceding_rule.destination_addresses:
        return True, "Source addresses are the same"

    if AnyObj in preceding_rule.destination_addresses:
        return True, "Preceding rule allows any source address"

    if AnyObj in rule.destination_addresses:
        return False, "Rule not covered due to 'any' source"

    for addr in rule.destination_addresses_ip:
        if not any(
            addr.subnet_of(net)
            for net in preceding_rule.destination_addresses_ip
        ):
            return (
                False,
                f"Source ip address {addr} is not covered by preceding rule",
            )

    return True, "Preceding rule covers all source ip addresses"


class ShadowingByValue(Shadowing):
    name = "Shadowing with addresses"
    checks: list[ShadowingCheckFunction] = [
        check_action,
        check_application,
        check_services,
        check_source_zone,
        check_destination_zone,
        check_source_addresses_by_ip,
        check_destination_addresses_by_ip,
    ]

    def __init__(
        self,
        security_rules: list["SecurityRule"],
        address_objects: list["AddressObject"],
        address_groups: list["AddressGroup"],
        lookup_class: type = AddressResolver,
    ):
        self.address_objects = address_objects
        self.address_groups = address_groups
        self.resolver = lookup_class(address_objects, address_groups)
        super().__init__(security_rules)

    def execute(self) -> dict[str, PrecedingRulesOutputs]:
        self.resolve_addresses()
        return super().execute()

    def resolve_addresses(
        self,
    ) -> None:
        """Resolve Security Rules ``source_addresses`` and ``destination_addresses`` values
        of Address Objects and Address Groups to an actual IP addresses.
        """
        logger.info(
            "↺ Resolving Address Groups and Address Objects actual IP addresses"
        )
        errors = []
        attr_pairs = (
            ("source_addresses", "source_addresses_ip"),
            ("destination_addresses", "destination_addresses_ip"),
        )
        for rule in self.security_rules:
            for origin_attr, desire_attr in attr_pairs:
                current_value = getattr(rule, origin_attr)
                if not current_value or AnyObj in current_value:
                    continue
                try:
                    resolved_value = self.resolver.resolve(current_value)
                    if resolved_value:
                        setattr(rule, desire_attr, resolved_value)
                except ValueError as ex:
                    logger.debug(ex)
                    errors.append(
                        f"rule={rule.name} {origin_attr}={current_value}"
                    )

        if errors:
            raise ValueError(
                f"Failed to resolve rules addresses: {' | '.join(errors)}",
            )



import logging
from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar

if TYPE_CHECKING:
    from policy_inspector.models import SecurityRule

logger = logging.getLogger(__name__)

ScenarioResults = TypeVar("ScenarioResults")

CheckResult = tuple[bool, str]
"""
A tuple representing the result of a check function.

1. ``bool``: Indicates whether the check was fulfilled or not.
2. ``str``: A verbose message describing the result.
"""

Check = Callable[[...], CheckResult]
"""A callable type definition for a scenario check function."""


class Scenario:
    """
    Base class for defining security scenarios and checks.

    Attributes:
        _scenarios: A set of all registered subclasses of Scenario.
        checks: A list of callable check functions to be executed on security rules.
    """

    name: Optional[str] = None
    checks: list[Check] = []

    _scenarios: dict[str, type["Scenario"]] = {}

    def __init_subclass__(cls, **kwargs) -> None:
        """Registers subclasses automatically in the `scenarios` set."""
        super().__init_subclass__(**kwargs)
        cls._scenarios[str(cls)] = cls

    def __str__(self):
        return self.name or self.__class__.__name__

    @classmethod
    def get_available(cls) -> dict[str, type["Scenario"]]:
        """
        Retrieve all registered ``Scenario`` subclasses.

        Returns:
            A set containing all subclasses of ``Scenario``.
        """
        return cls._scenarios

    @classmethod
    def get_by_name(cls, name: str) -> type["Scenario"]:
        return cls._scenarios[name]

    def execute(self) -> ScenarioResults:
        """
        Execute the scenario logic.

        Warnings:
            This method must be implemented by subclasses.

        Returns:
            The results of executing.
        """
        raise NotImplementedError

    def analyze(self, results: ScenarioResults) -> Any:
        """
        Analyze the results obtained from executing a scenario.

        Warnings:
            This method must be implemented by subclasses.

        Args:
            results: The results to analyze.

        Returns:
            The analysis outcome.
        """
        raise NotImplementedError

    def run_checks(self, *rules: "SecurityRule") -> dict[str, CheckResult]:
        """
        Run all defined ``checks`` against the provided security rule or rules.

        Args:
            *rules: Security rules to evaluate.

        Returns:
            A dictionary mapping check function names to their results (status and message).

        Logs exceptions if any check raises an error during execution.
        """
        results = {}
        for check in self.checks:
            try:
                results[check.__name__] = check(*rules)
            except Exception as ex:  # noqa: BLE001
                logger.warning(f"☠ Error: {ex}")
                logger.warning(f"☠ Check function: '{check.__name__}'")
                for i, rule in enumerate(rules, start=1):
                    logger.warning(f"☠ Rule {i}: {rule.name}")
                    logger.debug(f"☠ Rule {i}: {rule.model_dump()}")
        return results
